"""
Roman pipeline step for image alignment.
"""

import os
from pathlib import Path

import numpy as np
from astropy.table import Table
from roman_datamodels import datamodels as rdm
from stcal.tweakreg import tweakreg
from stcal.tweakreg.tweakreg import _SINGLE_GROUP_REFCAT_STR, SINGLE_GROUP_REFCAT

from romancal.assign_wcs.utils import add_s_region

# LOCAL
from ..datamodels import ModelLibrary
from ..stpipe import RomanStep

DEFAULT_ABS_REFCAT = SINGLE_GROUP_REFCAT[0]

__all__ = ["TweakRegStep"]


class TweakRegStep(RomanStep):
    """
    TweakRegStep: Image alignment based on catalogs of sources detected in
    input images.
    """

    class_alias = "tweakreg"

    spec = f"""
        use_custom_catalogs = boolean(default=False) # Use custom user-provided catalogs?
        catalog_format = string(default='ascii.ecsv') # Catalog output file format
        catfile = string(default='') # Name of the file with a list of custom user-provided catalogs
        catalog_path = string(default='') # Catalog output file path
        enforce_user_order = boolean(default=False) # Align images in user specified order?
        expand_refcat = boolean(default=False) # Expand reference catalog with new sources?
        minobj = integer(default=15) # Minimum number of objects acceptable for matching
        searchrad = float(default=2.0) # The search radius in arcsec for a match
        use2dhist = boolean(default=True) # Use 2d histogram to find initial offset?
        separation = float(default=1.0) # Minimum object separation in arcsec
        tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec
        fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift') # Fitting geometry
        nclip = integer(min=0, default=3) # Number of clipping iterations in fit
        sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units
        abs_refcat = string(default='{DEFAULT_ABS_REFCAT}')  # Absolute reference
        # catalog. Options: {_SINGLE_GROUP_REFCAT_STR}
        save_abs_catalog = boolean(default=False)  # Write out used absolute astrometric reference catalog as a separate product
        abs_minobj = integer(default=15) # Minimum number of objects acceptable for matching when performing absolute astrometry
        abs_searchrad = float(default=6.0) # The search radius in arcsec for a match when performing absolute astrometry
        # We encourage setting this parameter to True. Otherwise, xoffset and yoffset will be set to zero.
        abs_use2dhist = boolean(default=True) # Use 2D histogram to find initial offset when performing absolute astrometry?
        abs_separation = float(default=1.0) # Minimum object separation in arcsec when performing absolute astrometry
        abs_tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec when performing absolute astrometry
        # Fitting geometry when performing absolute astrometry
        abs_fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift')
        abs_nclip = integer(min=0, default=3) # Number of clipping iterations in fit when performing absolute astrometry
        abs_sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units when performing absolute astrometry
        output_use_model = boolean(default=True)  # When saving use `DataModel.meta.filename`
        update_source_catalog_coordinates = boolean(default=False) # Update source catalog file with tweaked coordinates?
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):

        # properly handle input
        try:
            if isinstance(input, rdm.DataModel):
                images = ModelLibrary([input])
            elif str(input).endswith(".asdf"):
                images = ModelLibrary([rdm.open(input)])
            elif isinstance(input, ModelLibrary):
                images = input
            else:
                images = ModelLibrary(input)
        except TypeError as e:
            e.args = (
                "Input to tweakreg must be a list of DataModels, an "
                "association, or an already open ModelLibrary "
                "containing one or more DataModels.",
            ) + e.args[1:]
            raise e

        if not images:
            raise ValueError("Input must contain at least one image model.")

        self.log.info(
            f"Number of image groups to be aligned: {len(images.group_indices):d}."
        )
        self.log.info("Image groups:")
        for name in images.group_names:
            self.log.info(f"  {name}")
        # set the first image as reference
        with images:
            ref_image = images.borrow(0)
            images.shelve(ref_image, 0, modify=False)

        catdict = _parse_catfile(self.catfile)

        use_custom_catalogs = self.use_custom_catalogs
        # if user requested the use of custom catalogs and provided a
        # valid 'catfile' file name that has no custom catalogs,
        # turn off the use of custom catalogs:
        if catdict is not None and not catdict:
            self.log.warning(
                "'use_custom_catalogs' is set to True but 'catfile' "
                "contains no user catalogs."
            )
            use_custom_catalogs = False

        if use_custom_catalogs and catdict:
            with images:
                for i, member in enumerate(images.asn["products"][0]["members"]):
                    filename = member["expname"]
                    if filename in catdict:
                        # FIXME: I'm not sure if this captures all the possible combinations
                        # for example, meta.tweakreg_catalog is set by the container (when
                        # it's present in the association). However the code in this step
                        # checks meta.source_catalog.tweakreg_catalog. I think this means
                        # that setting a catalog via an association does not work. Is this
                        # intended? If so, the container can be updated to not support that.
                        model = images.borrow(i)
                        model.meta["source_detection"] = {
                            "tweakreg_catalog_name": catdict[filename],
                        }
                        images.shelve(model, i)
                    else:
                        images.shelve(model, i, modify=False)

        # set path where the source catalog will be saved to
        if len(self.catalog_path) == 0:
            self.catalog_path = os.getcwd()
        self.catalog_path = Path(self.catalog_path).as_posix()
        self.log.info(f"All source catalogs will be saved to: {self.catalog_path}")

        # set reference catalog name
        if not self.abs_refcat:
            self.abs_refcat = DEFAULT_ABS_REFCAT.strip().upper()
        if self.abs_refcat != DEFAULT_ABS_REFCAT:
            self.expand_refcat = True

        # build the catalogs for input images
        imcats = []
        with images:
            for i, image_model in enumerate(images):
                exposure_type = image_model.meta.exposure.type
                if exposure_type != "WFI_IMAGE":
                    self.log.info("Skipping TweakReg for spectral exposure.")
                    image_model.meta.cal_step.tweakreg = "SKIPPED"
                else:
                    source_detection = getattr(
                        image_model.meta, "source_detection", None
                    )
                    if source_detection is None:
                        images.shelve(image_model, i, modify=False)
                        raise AttributeError(
                            "Attribute 'meta.source_detection' is missing. "
                            "Please either run SourceDetectionStep or provide a custom source catalog."
                        )

                    try:
                        catalog = self.get_tweakreg_catalog(
                            source_detection, image_model
                        )
                    except AttributeError as e:
                        self.log.error(f"Failed to retrieve tweakreg_catalog: {e}")
                        images.shelve(image_model, i, modify=False)
                        raise e

                    try:
                        # validate catalog columns
                        _validate_catalog_columns(catalog)
                    except ValueError as e:
                        self.log.error(f"Failed to validate catalog columns: {e}")
                        images.shelve(image_model, i, modify=False)
                        raise e

                    catalog = tweakreg.filter_catalog_by_bounding_box(
                        catalog, image_model.meta.wcs.bounding_box
                    )

                    if self.save_abs_catalog:
                        output_name = os.path.join(
                            self.catalog_path, f"fit_{self.abs_refcat.lower()}_ref.ecsv"
                        )
                        catalog.write(
                            output_name, format=self.catalog_format, overwrite=True
                        )

                    image_model.meta["tweakreg_catalog"] = catalog.as_array()
                    nsources = len(catalog)
                    self.log.info(
                        f"Detected {nsources} sources in {image_model.meta.filename}."
                        if nsources
                        else f"No sources found in {image_model.meta.filename}."
                    )
                    # build image catalog
                    # catalog name
                    catalog_name = os.path.splitext(image_model.meta.filename)[0].strip(
                        "_- "
                    )
                    # catalog data
                    catalog_table = Table(image_model.meta.tweakreg_catalog)
                    catalog_table.meta["name"] = catalog_name

                    imcat = tweakreg.construct_wcs_corrector(
                        wcs=image_model.meta.wcs,
                        refang=image_model.meta.wcsinfo,
                        catalog=catalog_table,
                        group_id=image_model.meta.group_id,
                    )
                    imcat.meta["model_index"] = i
                    imcats.append(imcat)
                images.shelve(image_model, i)

        # run alignment only if it was possible to build image catalogs
        if len(imcats):
            # extract WCS correctors to use for image alignment
            if len(images.group_indices) > 1:
                self.do_relative_alignment(imcats)

            if self.abs_refcat in SINGLE_GROUP_REFCAT:
                self.do_absolute_alignment(ref_image, imcats)

            # finalize step
            with images:
                for imcat in imcats:
                    image_model = images.borrow(imcat.meta["model_index"])
                    image_model.meta.cal_step["tweakreg"] = "COMPLETE"
                    # remove source catalog
                    del image_model.meta["tweakreg_catalog"]

                    # retrieve fit status and update wcs if fit is successful:
                    if "SUCCESS" in imcat.meta.get("fit_info")["status"]:
                        # Update/create the WCS .name attribute with information
                        # on this astrometric fit as the only record that it was
                        # successful:

                        # NOTE: This .name attrib agreed upon by the JWST Cal
                        #       Working Group.
                        #       Current value is merely a place-holder based
                        #       on HST conventions. This value should also be
                        #       translated to the FITS WCSNAME keyword
                        #       IF that is what gets recorded in the archive
                        #       for end-user searches.
                        imcat.wcs.name = f"FIT-LVL2-{self.abs_refcat}"

                        # serialize object from tweakwcs
                        # (typecasting numpy objects to python types so that it doesn't cause an
                        # issue when saving datamodel to ASDF)
                        wcs_fit_results = {
                            k: (
                                v.tolist()
                                if isinstance(v, (np.ndarray, np.bool_))
                                else v
                            )
                            for k, v in imcat.meta["fit_info"].items()
                        }
                        # add fit results and new WCS to datamodel
                        image_model.meta["wcs_fit_results"] = wcs_fit_results
                        # remove unwanted keys from WCS fit results
                        for k in [
                            "eff_minobj",
                            "matched_ref_idx",
                            "matched_input_idx",
                            "fit_RA",
                            "fit_DEC",
                            "fitmask",
                        ]:
                            del image_model.meta["wcs_fit_results"][k]

                        # update WCS
                        image_model.meta.wcs = imcat.wcs
                        # update S_REGION
                        add_s_region(image_model)

                    images.shelve(image_model, imcat.meta["model_index"])

        return images

    def update_catalog_coordinates(self, tweakreg_catalog_name, tweaked_wcs):
        """
        Update the source catalog coordinates using the tweaked WCS.

        Parameters
        ----------
        tweakreg_catalog_name : str
            The name of the TweakReg catalog file produced by `SourceCatalog`.
        tweaked_wcs : `gwcs.wcs.WCS`
            The tweaked World Coordinate System (WCS) object.

        Returns
        -------
        None
        """
        # read in cat file
        with rdm.open(tweakreg_catalog_name) as source_catalog_model:
            # get catalog
            catalog = source_catalog_model.source_catalog

            # define mapping between pixel and world coordinates
            colname_mapping = {
                ("xcentroid", "ycentroid"): ("ra_centroid", "dec_centroid"),
                ("x_psf", "y_psf"): ("ra_psf", "dec_psf"),
            }

            for k, v in colname_mapping.items():
                # get column names
                x_colname, y_colname = k
                ra_colname, dec_colname = v

                # calculate new coordinates using tweaked WCS and update catalog coordinates
                catalog[ra_colname], catalog[dec_colname] = tweaked_wcs(
                    catalog[x_colname], catalog[y_colname]
                )

            # save updated catalog (overwrite cat file)
            self.save_model(
                source_catalog_model,
                output_file=source_catalog_model.meta.filename,
                suffix="cat",
                force=True,
            )

    def read_catalog(self, catalog_name):
        """
        Reads a source catalog from a specified file.

        This function determines the format of the catalog based on the file extension.
        If the file ends with "asdf", it uses a specific method to open and read the catalog;
        otherwise, it reads the catalog using a standard table format.

        Parameters
        ----------
        catalog_name : str
            The name of the catalog file to read.

        Returns
        -------
        Table
            The read catalog as a Table object.

        Raises
        ------
        ValueError
            If the catalog format is unsupported.
        """
        if catalog_name.endswith("asdf"):
            with rdm.open(catalog_name) as source_catalog_model:
                catalog = source_catalog_model.source_catalog
        else:
            catalog = Table.read(catalog_name, format=self.catalog_format)
        return catalog

    def get_tweakreg_catalog(self, source_detection, image_model):
        """
        Retrieve the tweakreg catalog from source detection.

        This method checks the source detection metadata for the presence of a
        tweakreg catalog data or a string with its name. It returns the catalog
        as a Table object if either is found, or raises an error if neither is available.

        Parameters
        ----------
        source_detection : object
            The source detection metadata containing catalog information.
        image_model : DataModel
            The image model associated with the source detection.

        Returns
        -------
        Table
            The retrieved tweakreg catalog as a Table object.

        Raises
        ------
        AttributeError
            If the required catalog information is missing from the source detection.
        """
        if getattr(source_detection, "tweakreg_catalog", None):
            tweakreg_catalog = Table(np.asarray(source_detection.tweakreg_catalog))
            del image_model.meta.source_detection["tweakreg_catalog"]
            return tweakreg_catalog

        if getattr(source_detection, "tweakreg_catalog_name", None):
            return self.read_catalog(source_detection.tweakreg_catalog_name)

        raise AttributeError(
            "Attribute 'meta.source_detection.tweakreg_catalog' is missing. "
            "Please either run SourceDetectionStep or provide a custom source catalog."
        )

    def do_relative_alignment(self, imcats):
        """
        Perform relative alignment of images.

        This method performs relative alignment with the specified parameters,
        including search radius, separation, and fitting geometry.

        Parameters
        ----------
        imcats : list
            A list of image catalogs containing source information for alignment.

        Returns
        -------
        None
        """
        tweakreg.relative_align(
            imcats,
            searchrad=self.searchrad,
            separation=self.separation,
            use2dhist=self.use2dhist,
            tolerance=self.tolerance,
            xoffset=0,
            yoffset=0,
            enforce_user_order=self.enforce_user_order,
            expand_refcat=self.expand_refcat,
            minobj=self.minobj,
            fitgeometry=self.fitgeometry,
            nclip=self.nclip,
            sigma=self.sigma,
            clip_accum=True,
        )

    def do_absolute_alignment(self, ref_image, imcats):
        """
        Perform absolute alignment of images.

        This method retrieves a reference image and performs absolute alignment
        using the specified parameters, including reference WCS information and
        catalog details. It aligns the provided image catalogs to the absolute
        reference catalog.

        Parameters
        ----------
        ref_image : DataModel
            The reference image used for alignment, which contains WCS information.
        imcats : list
            A list of image catalogs containing source information for alignment.

        Returns
        -------
        None
        """
        tweakreg.absolute_align(
            imcats,
            self.abs_refcat,
            ref_wcs=ref_image.meta.wcs,
            ref_wcsinfo=ref_image.meta.wcsinfo,
            epoch=ref_image.meta.exposure.mid_time.decimalyear,
            abs_minobj=self.abs_minobj,
            abs_fitgeometry=self.abs_fitgeometry,
            abs_nclip=self.abs_nclip,
            abs_sigma=self.abs_sigma,
            abs_searchrad=self.abs_searchrad,
            abs_use2dhist=self.abs_use2dhist,
            abs_separation=self.abs_separation,
            abs_tolerance=self.abs_tolerance,
            save_abs_catalog=self.save_abs_catalog,
            abs_catalog_output_dir=self.output_dir,
            clip_accum=True,
        )


def _parse_catfile(catfile):
    """
    Parse a catalog file and return a dictionary mapping data models to catalog paths.

    This function reads a specified catalog file, extracting data model names and
    their associated catalog paths. It supports a format where each line contains
    a data model followed by an optional catalog path, and it ensures that the
    file adheres to the expected structure.

    Parameters
    ----------
    catfile : str
        The path to the catalog file to be parsed.

    Returns
    -------
    dict or None
        A dictionary mapping data model names to catalog paths, or None if the
        input file is empty or invalid.

    Raises
    ------
    ValueError
        If the catalog file contains more than two columns per line.
    """

    if catfile is None or not catfile.strip():
        return None

    catdict = {}

    with open(catfile) as f:
        catfile_dir = os.path.dirname(catfile)

        for line in f:
            sline = line.strip()
            if not sline or sline[0] == "#":
                continue

            data_model, *catalog = sline.split()
            catalog = list(map(str.strip, catalog))
            if len(catalog) == 1:
                catdict[data_model] = os.path.join(catfile_dir, catalog[0])
            elif not catalog:
                catdict[data_model] = None
            else:
                raise ValueError("'catfile' can contain at most two columns.")

    return catdict


def _validate_catalog_columns(catalog):
    """
    Validate the presence of required columns in the catalog.

    This method checks if the specified axis column exists in the catalog.
    If the axis is not found, it looks for a corresponding centroid column
    and renames it if present. If neither is found, it raises an error.

    Parameters
    ----------
    catalog : Table
        The catalog to validate, which should contain source information.
    axis : str
        The axis to check for in the catalog (e.g., 'x' or 'y').

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the required columns are missing from the catalog.
    """
    for axis in ["x", "y"]:
        if axis not in catalog.colnames:
            long_axis = f"{axis}centroid"
            if long_axis in catalog.colnames:
                catalog.rename_column(long_axis, axis)
            else:
                raise ValueError(
                    "'tweakreg' source catalogs must contain a header with "
                    "columns named either 'x' and 'y' or 'xcentroid' and 'ycentroid'."
                )
    return catalog



import os
from pathlib import Path

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
from numpy.testing import assert_equal
from photutils.segmentation import SegmentationImage
from roman_datamodels import datamodels as rdm
from roman_datamodels.datamodels import (
    ImageModel,
    MosaicModel,
    MosaicSegmentationMapModel,
    MosaicSourceCatalogModel,
    SegmentationMapModel,
    SourceCatalogModel,
)
from roman_datamodels.maker_utils import mk_level2_image, mk_level3_mosaic

from romancal.source_catalog.reference_data import ReferenceData
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.source_catalog.source_catalog_step import SourceCatalogStep


def make_test_image():
    g1 = Gaussian2D(121.0, 11.1, 12.2, 1.5, 1.5)
    g2 = Gaussian2D(70, 65, 18, 9.2, 4.5)
    g3 = Gaussian2D(111.0, 41, 42.7, 8.0, 3.0, theta=30 * u.deg)
    g4 = Gaussian2D(81.0, 17, 52.7, 4, 2, theta=102 * u.deg)
    g5 = Gaussian2D(107.0, 65, 71, 12, 2, theta=142 * u.deg)
    g6 = Gaussian2D(50, 20, 80, 2.1, 2.1)
    g7 = Gaussian2D(97.0, 85, 87.3, 4, 2, theta=-30 * u.deg)

    yy, xx = np.mgrid[0:101, 0:101]
    data = (
        g1(xx, yy)
        + g2(xx, yy)
        + g3(xx, yy)
        + g4(xx, yy)
        + g5(xx, yy)
        + g6(xx, yy)
        + g7(xx, yy)
    ).value.astype("float32")

    y0 = 2
    x0 = 90
    dd = 5
    value = -20
    y1 = y0 + dd
    x1 = x0 + dd - 1
    x2 = x0 + dd
    x3 = x0 + 2 * dd - 1
    data[y0:y1, x0] = value
    data[y0, x0:x1] = value
    data[y0:y1, x2] = value
    data[y0, x2:x3] = value
    data[y0 + dd // 2, x2:x3] = value
    data[y1 - 1, x2:x3] = value
    data[y0 + 1 : y1 - 1, x3] = value

    rng = np.random.default_rng(seed=123)
    noise = rng.normal(0, 2.5, size=data.shape)
    data += noise
    err = data / 10.0

    return data, err


@pytest.fixture
def mosaic_model():
    wfi_mosaic = mk_level3_mosaic(shape=(101, 101))
    model = MosaicModel(wfi_mosaic)
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.weight = 1.0 / err
    return model


@pytest.fixture
def image_model():
    wfi_image = mk_level2_image(shape=(101, 101))
    model = ImageModel(wfi_image)
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.meta.photometry.conversion_megajanskys = (0.3324 * u.MJy / u.sr).value
    return model


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results",
    (
        (3, 10, 7, True),
        (3, 50, 5, False),
        (10, 10, 7, False),
        (20, 10, 5, False),
        (25, 10, 3, False),
        (35, 10, 1, False),
        (50, 10, 0, False),
    ),
)
def test_l2_source_catalog(
    image_model, snr_threshold, npixels, nsources, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)

    columns = [
        "label",
        "xcentroid",
        "ycentroid",
        "ra_centroid",
        "dec_centroid",
        "aper_bkg_flux",
        "aper_bkg_flux_err",
        "aper30_flux",
        "aper30_flux_err",
        "aper50_flux",
        "aper50_flux_err",
        "aper70_flux",
        "aper70_flux_err",
        "aper_total_flux",
        "aper_total_flux_err",
        "CI_50_30",
        "CI_70_50",
        "CI_70_30",
        "is_extended",
        "sharpness",
        "roundness",
        "nn_label",
        "nn_dist",
        "isophotal_flux",
        "isophotal_flux_err",
        "isophotal_area",
        "semimajor_sigma",
        "semiminor_sigma",
        "ellipticity",
        "orientation",
        "sky_orientation",
        "ra_bbox_ll",
        "dec_bbox_ll",
        "ra_bbox_ul",
        "dec_bbox_ul",
        "ra_bbox_lr",
        "dec_bbox_lr",
        "ra_bbox_ur",
        "dec_bbox_ur",
    ]

    assert len(cat) == nsources

    if nsources > 0:
        for col in columns:
            assert col in cat.colnames
        assert np.min(cat["xcentroid"]) > 0.0
        assert np.min(cat["ycentroid"]) > 0.0
        assert np.max(cat["xcentroid"]) < 100.0
        assert np.max(cat["ycentroid"]) < 100.0


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results",
    (
        (3, 10, 7, True),
        (3, 50, 5, False),
        (10, 10, 7, False),
        (20, 10, 5, False),
        (25, 10, 3, False),
        (35, 10, 1, False),
        (50, 10, 0, False),
    ),
)
def test_l3_source_catalog(
    mosaic_model, snr_threshold, npixels, nsources, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = SourceCatalogStep()

    result = step.call(
        mosaic_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)

    columns = [
        "label",
        "xcentroid",
        "ycentroid",
        "ra_centroid",
        "dec_centroid",
        "aper_bkg_flux",
        "aper_bkg_flux_err",
        "aper30_flux",
        "aper30_flux_err",
        "aper50_flux",
        "aper50_flux_err",
        "aper70_flux",
        "aper70_flux_err",
        "aper_total_flux",
        "aper_total_flux_err",
        "CI_50_30",
        "CI_70_50",
        "CI_70_30",
        "is_extended",
        "sharpness",
        "roundness",
        "nn_label",
        "nn_dist",
        "isophotal_flux",
        "isophotal_flux_err",
        "isophotal_area",
        "semimajor_sigma",
        "semiminor_sigma",
        "ellipticity",
        "orientation",
        "sky_orientation",
        "ra_bbox_ll",
        "dec_bbox_ll",
        "ra_bbox_ul",
        "dec_bbox_ul",
        "ra_bbox_lr",
        "dec_bbox_lr",
        "ra_bbox_ur",
        "dec_bbox_ur",
    ]

    assert len(cat) == nsources

    if nsources > 0:
        for col in columns:
            assert col in cat.colnames
        assert np.min(cat["xcentroid"]) > 0.0
        assert np.min(cat["ycentroid"]) > 0.0
        assert np.max(cat["xcentroid"]) < 100.0
        assert np.max(cat["ycentroid"]) < 100.0


@pytest.mark.webbpsf
def test_background(mosaic_model, tmp_path):
    """
    Test background fallback when Background2D fails.
    """
    os.chdir(tmp_path)
    step = SourceCatalogStep()
    result = step.call(
        mosaic_model,
        bkg_boxsize=1000,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=25,
        fit_psf=False,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)


@pytest.mark.webbpsf
def test_l2_input_model_unchanged(image_model, tmp_path):
    """
    Test that the input model data and error arrays are unchanged after
    processing by SourceCatalogStep.
    """
    os.chdir(tmp_path)
    original_data = image_model.data.copy()
    original_err = image_model.err.copy()

    step = SourceCatalogStep()
    step.call(
        image_model,
        snr_threshold=0.5,
        npixels=5,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        save_results=False,
    )

    assert_equal(original_data, image_model.data)
    assert_equal(original_err, image_model.err)


@pytest.mark.webbpsf
def test_l3_input_model_unchanged(mosaic_model, tmp_path):
    """
    Test that the input model data and error arrays are unchanged after
    processing by SourceCatalogStep.
    """
    os.chdir(tmp_path)
    original_data = mosaic_model.data.copy()
    original_err = mosaic_model.err.copy()

    step = SourceCatalogStep()
    step.call(
        mosaic_model,
        snr_threshold=0.5,
        npixels=5,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        save_results=False,
    )

    assert_equal(original_data, mosaic_model.data)
    assert_equal(original_err, mosaic_model.err)


@pytest.mark.webbpsf
def test_inputs(mosaic_model):
    with pytest.raises(ValueError):
        ReferenceData(np.ones((3, 3)), (30, 50, 70))
    with pytest.raises(ValueError):
        aperture_ee = (70, 50, 30)
        ReferenceData(mosaic_model, aperture_ee)
    with pytest.raises(ValueError):
        aperture_ee = (30, 50)
        ReferenceData(mosaic_model, aperture_ee)
    with pytest.raises(ValueError):
        aperture_ee = (-1, 50, 70)
        ReferenceData(mosaic_model, aperture_ee)
    with pytest.raises(ValueError):
        aperture_ee = (40, 70, 150)
        ReferenceData(mosaic_model, aperture_ee)

    data = np.ones((3, 3), dtype=int)
    data[1, 1] = 1
    segm = SegmentationImage(data)
    cdata = np.ones((3, 3))
    aper_params = {}
    ci_thresh = 100.0
    with pytest.raises(ValueError):
        RomanSourceCatalog(
            np.ones((3, 3)), segm, cdata, aper_params, ci_thresh, 2.0, fit_psf=True
        )

    with pytest.raises(ValueError):
        RomanSourceCatalog(
            mosaic_model, segm, cdata, aper_params, (1.0, 2.0, 3.0), 2.0, fit_psf=True
        )


@pytest.mark.webbpsf
def test_do_psf_photometry(tmp_path, image_model):
    """
    Test that do_psf_photometry can recover mock sources and their position and photometry.
    """
    os.chdir(tmp_path)

    # get column names mapping for PSF photometry
    psf_colnames_mapping = (
        RomanSourceCatalog.get_psf_photometry_catalog_colnames_mapping()
    )
    psf_colnames = [
        x.get("new_name")
        for x in psf_colnames_mapping
        if x.get("old_name") in ["x_fit", "y_fit", "flux_fit"]
    ]

    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=20,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
        save_results=False,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)

    # check the number of sources that have been detected
    assert len(cat) == 7
    # check that all sources have both position and flux determined (ignore errors/flags)
    for col_name in psf_colnames:
        assert len(cat[col_name])  # make sure the column isn't empty
        assert not np.any(np.isnan(cat[col_name]))  # and contains no nans


@pytest.mark.webbpsf
@pytest.mark.parametrize("fit_psf", [True, False])
def test_do_psf_photometry_column_names(tmp_path, image_model, fit_psf):
    """
    Test that fit_psf will determine whether the PSF
    photometry columns are added to the final catalog or not.
    """
    os.chdir(tmp_path)

    # get column names mapping for PSF photometry
    psf_colnames_mapping = (
        RomanSourceCatalog.get_psf_photometry_catalog_colnames_mapping()
    )

    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=20,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
        save_results=False,
        fit_psf=fit_psf,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)

    # check if the PSF photometry column names are present or not based on fit_psf value
    psf_colnames_present = all(
        x.get("new_name") in cat.colnames for x in psf_colnames_mapping
    )
    psf_colnames_not_present = all(
        x.get("new_name") not in cat.colnames for x in psf_colnames_mapping
    )

    assert (fit_psf and psf_colnames_present) or (
        not fit_psf and psf_colnames_not_present
    )


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results, return_updated_model, expected_result, expected_outputs",
    (
        (
            3,
            10,
            7,
            True,
            True,
            ImageModel,
            {
                "cat": SourceCatalogModel,
                "segm": SegmentationMapModel,
                "sourcecatalog": ImageModel,
            },
        ),
        (
            3,
            50,
            5,
            True,
            False,
            SourceCatalogModel,
            {
                "cat": SourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
        (
            10,
            10,
            7,
            False,
            True,
            ImageModel,
            {
                "cat": SourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
        (
            20,
            10,
            5,
            False,
            False,
            SourceCatalogModel,
            {
                "cat": SourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
    ),
)
def test_l2_source_catalog_keywords(
    image_model,
    snr_threshold,
    npixels,
    nsources,
    save_results,
    return_updated_model,
    expected_result,
    expected_outputs,
    tmp_path,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep
    and that the desired output files are saved to the disk with the correct type.
    """
    os.chdir(tmp_path)
    step = SourceCatalogStep
    # this step attribute controls whether to return a datamodel or source catalog
    step.return_updated_model = return_updated_model

    result = step.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    # assert that we returned the correct object
    assert isinstance(result, expected_result)

    # assert that the desired output files were saved to disk
    assert all(
        Path(tmp_path / f"{result.meta.filename}_{suffix}.asdf").exists()
        for suffix in expected_outputs.keys()
    )

    # assert that the desired output files were saved with the correct datamodel type
    assert all(
        isinstance(
            rdm.open(Path(tmp_path / f"{result.meta.filename}_{suffix}.asdf")),
            expected_outputs.get(suffix),
        )
        for suffix in expected_outputs.keys()
    )


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results, return_updated_model, expected_result, expected_outputs",
    (
        (
            3,
            10,
            7,
            True,
            True,
            MosaicModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
                "sourcecatalog": MosaicModel,
            },
        ),
        (
            3,
            50,
            5,
            True,
            False,
            MosaicSourceCatalogModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
        (
            10,
            10,
            7,
            False,
            True,
            MosaicModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
        (
            20,
            10,
            5,
            False,
            False,
            MosaicSourceCatalogModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
    ),
)
def test_l3_source_catalog_keywords(
    mosaic_model,
    snr_threshold,
    npixels,
    nsources,
    save_results,
    return_updated_model,
    expected_result,
    expected_outputs,
    tmp_path,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep
    and that the desired output files are saved to the disk with the correct type.
    """
    os.chdir(tmp_path)
    step = SourceCatalogStep
    # this step attribute controls whether to return a datamodel or source catalog
    step.return_updated_model = return_updated_model

    result = step.call(
        mosaic_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    # assert that we returned the correct object
    assert isinstance(result, expected_result)

    # assert that the desired output files were saved to disk
    assert all(
        Path(tmp_path / f"{result.meta.filename}_{suffix}.asdf").exists()
        for suffix in expected_outputs.keys()
    )

    # assert that the desired output files were saved with the correct datamodel type
    assert all(
        isinstance(
            rdm.open(Path(tmp_path / f"{result.meta.filename}_{suffix}.asdf")),
            expected_outputs.get(suffix),
        )
        for suffix in expected_outputs.keys()
    )


@pytest.mark.webbpsf
@pytest.mark.parametrize(
    "return_updated_model, expected_result",
    (
        (
            True,
            ImageModel,
        ),
        (
            False,
            SourceCatalogModel,
        ),
    ),
)
def test_l2_source_catalog_return_updated_model_attribute(
    image_model,
    return_updated_model,
    expected_result,
    tmp_path,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep.
    """
    os.chdir(tmp_path)

    step = SourceCatalogStep(
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
    )

    if return_updated_model:
        # mimic what happens in the ELP -- i.e. set the "hidden" parameter
        # to cause this step to return a model instead of a catalog
        step.return_updated_model = return_updated_model

    result = step.run(image_model)

    # assert that we returned the correct object
    assert isinstance(result, expected_result)



import shutil

import pytest

SCRIPTS = [
    "okify_regtests",
    "verify_install_requires",
]


@pytest.mark.parametrize("script", SCRIPTS)
def test_script_installed(script):
    assert shutil.which(script) is not None, f"`{script}` not installed"



import logging
import warnings

import numpy as np
from astropy.convolution import convolve_fft
from astropy.utils.exceptions import AstropyUserWarning
from photutils.segmentation import make_2dgaussian_kernel

from romancal.datamodels import ModelLibrary

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def make_gaussian_kernel(fwhm):
    """
    Make a normalized 2D circular Gaussian kernel.

    The kernel will have odd sizes in both X and Y, be centered in the
    central pixel, and normalized to sum to 1.

    Parameters
    ----------
    fwhm : float
        The full-width at half-maximum (FWHM) in pixels of the 2D
        Gaussian kernel.

    Returns
    -------
    kernel : `astropy.convolution.Kernel2D`
        The output Gaussian kernel.
    """
    size = np.ceil(3 * fwhm).astype(int)
    size = size + 1 if size % 2 == 0 else size  # make size be odd
    kernel = make_2dgaussian_kernel(fwhm, size=size)  # sums to 1
    return kernel.array


def make_det_image(library, kernel_fwhm):
    """
    Make a detection image from a library of models.

    The detection image is the weighted sum of the input models, where
    the weights are SED weight divided by the read noise variance.

    Parameters
    ----------
    library : `romancal.datamodels.ModelLibrary`
        The input library of models.

    kernel_fwhm : float
        The full-width at half-maximum (FWHM) in pixels of the 2D
        Gaussian kernel used to smooth the detection image.

    Returns
    -------
    detection_data : 2D `numpy.ndarray`
        The detection image data.

    detection_error : 2D `numpy.ndarray`
        The detection image (standard deviation) error.
    """
    if not isinstance(library, ModelLibrary):
        raise TypeError("library input must be a ModelLibrary object")

    # TODO: extend to different kernel shapes beyond Gaussian?
    # would require different/additional kernel parameters
    kernel = make_gaussian_kernel(kernel_fwhm)

    log.info(f"Making detection image with kernel FWHM={kernel_fwhm}")

    with library:
        detection_data = 0.0
        detection_var = 0.0
        wht_sum = 0.0
        for i, model in enumerate(library):
            # TODO: SED weights to be defined in the asn file for each
            # input filter image
            try:
                sed_weight = library.asn["products"][0]["members"][i]["sed_weight"]
            except KeyError:
                sed_weight = 1.0

            log.info(
                f"Processing model {model.meta.filename}: "
                f"filter={model.meta.basic.optical_element}, {sed_weight=}"
            )

            # Ideally the weights should be the inverse variance of
            # all sources of noise except the source Poisson noise. The
            # closest approximation we have is var_rnoise (read noise
            # variance) -- var_rnoise is also used as the weighting in the
            # resample step. Note that model.err is the total error, which
            # includes source Poisson noise.
            wht = sed_weight / model.var_rnoise  # inverse variance
            wht_sum += wht

            coverage_mask = np.isnan(model.err)
            with warnings.catch_warnings():
                # Suppress warnings about any NaNs in the data
                warnings.filterwarnings(action="ignore", category=AstropyUserWarning)

                detection_data += convolve_fft(
                    wht * model.data, kernel, mask=coverage_mask
                )
                detection_var += convolve_fft(
                    wht**2 * model.var_rnoise,
                    kernel**2,
                    mask=coverage_mask,
                    normalize_kernel=False,
                    nan_treatment="fill",
                )

            library.shelve(model, modify=False)

    detection_data /= wht_sum
    detection_error = np.sqrt(detection_var) / wht_sum  # std dev error

    return detection_data, detection_error


def make_detection_image(library, kernel_fwhms):
    """
    Make a detection image from a library of models.

    The detection image is maximum of the detection images from each of
    the kernels.

    Parameters
    ----------
    library : `romancal.datamodels.ModelLibrary`
        The input library of models.

    kernel_fwhms : list of float
        The full-width at half-maximum (FWHM) in pixels of the 2D
        Gaussian kernels used to smooth the detection image.

    Returns
    -------
    detection_data : 2D `numpy.ndarray`
        The detection image data.

    detection_error : 2D `numpy.ndarray`
        The detection image (standard deviation) error.
    """
    log.info("Making detection image")
    det_img = 0
    det_err = 0
    for kernel_fwhm in kernel_fwhms:
        img, err = make_det_image(library, kernel_fwhm)
        det_img = np.maximum(det_img, img)
        det_err = np.maximum(det_err, err)
    return det_img, det_err



"""
Define the I/O methods for Level 3 associations
"""

import json as json_lib
import logging

import numpy as np
import yaml as yaml_lib

from .association import Association
from .exceptions import AssociationNotValidError
from .lib.member import Member

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = []


# Define JSON encoder to convert `Member` to `dict`
class AssociationEncoder(json_lib.JSONEncoder):
    """Encode to handle Associations"""

    def default(self, obj):
        # Convert Member to a simple dict
        if isinstance(obj, Member):
            return obj.data


@Association.ioregistry
class json:
    """Load and store associations as JSON"""

    @staticmethod
    def load(cls, serialized):
        """Unserialize an association from JSON

        Parameters
        ----------
        cls : class
            The class from which further information will be gathered
            and possibly instantiated.

        serialized : str or file object
            The JSON to read

        Returns
        -------
        association : dict
            The association

        Raises
        ------
        AssociationNotValidError
            Cannot create or validate the association.
        """
        if isinstance(serialized, str):
            loader = json_lib.loads
        else:
            # Presume a file object
            serialized.seek(0)
            loader = json_lib.load
        try:
            asn = loader(serialized)
        except Exception as err:
            logger.debug(f'Error unserializing: "{err}"')
            raise AssociationNotValidError(f'Container is not JSON: "{serialized}"')

        return asn

    @staticmethod
    def dump(asn):
        """Create JSON representation.

        Parameters
        ----------
        asn : Association
            The association to serialize

        Returns
        -------
        (name, str):
            Tuple where the first item is the suggested
            Name for the JSON file.
            Second item is the string containing the JSON serialization.
        """
        asn_filename = asn.asn_name
        if not asn_filename.endswith(".json"):
            asn_filename = asn_filename + ".json"
        return (
            asn_filename,
            json_lib.dumps(
                asn.data, cls=AssociationEncoder, indent=4, separators=(",", ": ")
            ),
        )


@Association.ioregistry
class yaml:
    """Load and store associations as YAML"""

    @staticmethod
    def load(cls, serialized):
        """Unserialize an association from YAML

        Parameters
        ----------
        cls : class
            The class from which further information will be gathered
            and possibly instantiated.

        serialized : str or file object
            The YAML to read

        Returns
        -------
        association : dict
            The association

        Raises
        ------
        AssociationNotValidError
            Cannot create or validate the association.
        """
        try:
            serialized.seek(0)
        except AttributeError:
            pass
        try:
            asn = yaml_lib.safe_load(serialized)
        except Exception as err:
            logger.debug(f'Error unserializing: "{err}"')
            raise AssociationNotValidError(f'Container is not YAML: "{serialized}"')
        return asn

    @staticmethod
    def dump(asn):
        """Create YAML representation.

         Parameters
        ----------
        asn : Association
            The association to serialize


        Returns
        -------
        (name, str):
            Tuple where the first item is the suggested
            Name for the YAML file.
            Second item is the string containing the YAML serialization.
        """
        asn_filename = asn.asn_name
        if not asn.asn_name.endswith(".yaml"):
            asn_filename = asn.asn_name + ".yaml"
        return (asn_filename, yaml_lib.dump(asn.data, default_flow_style=False))


# Register YAML representers
def np_str_representer(dumper, data):
    """Convert numpy.str_ into standard YAML string"""
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


yaml_lib.add_representer(np.str_, np_str_representer)


def member_representer(dumper, member):
    """Convert a Member to its basic dict representation"""
    return dumper.represent_dict(member.data)


yaml_lib.add_representer(Member, member_representer)



import os
from itertools import product

import astropy.units as u
import numpy as np
import pytest
from astropy import coordinates as coord
from astropy.modeling import models
from gwcs import coordinate_frames as cf
from gwcs import wcs as gwcs_wcs
from roman_datamodels.datamodels import ImageModel
from roman_datamodels.dqflags import pixel
from roman_datamodels.maker_utils import mk_level2_image, mk_sky_background

from romancal.datamodels import ModelLibrary
from romancal.skymatch import SkyMatchStep


def mk_gwcs(shape, sky_offset=[0, 0] * u.arcsec, rotate=0 * u.deg):
    # Example adapted from photutils:
    #   https://github.com/astropy/photutils/blob/
    #   2825356f1d876cacefb3a03d104a4c563065375f/photutils/datasets/make.py#L821
    rho = np.pi / 3.0
    # Roman plate scale:
    scale = (0.11 * u.arcsec / u.pixel).to_value(u.deg / u.pixel)

    shift_by_crpix = models.Shift((-shape[1] / 2) + 1) & models.Shift(
        (-shape[0] / 2) + 1
    )

    cd_matrix = np.array(
        [
            [-scale * np.cos(rho), scale * np.sin(rho)],
            [scale * np.sin(rho), scale * np.cos(rho)],
        ]
    )

    rotation = models.AffineTransformation2D(cd_matrix, translation=[0, 0])
    rotation.inverse = models.AffineTransformation2D(
        np.linalg.inv(cd_matrix), translation=[0, 0]
    )

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        197.8925 + sky_offset[0].to_value(u.deg),
        -1.36555556 + sky_offset[1].to_value(u.deg),
        180.0 + rotate.to_value(u.deg),
    )

    det2sky = shift_by_crpix | rotation | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )

    sky_frame = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]

    return gwcs_wcs.WCS(pipeline)


def mk_image_model(
    rate_mean=0,
    rate_std=1e-4,
    sky_offset=[0, 0] * u.arcsec,
    rotation=0 * u.deg,
    image_shape=(100, 100),
    rng=np.random.default_rng(619),
):
    l2 = mk_level2_image(shape=image_shape)
    l2_im = ImageModel(l2)
    l2_im.data = rng.normal(
        loc=rate_mean, scale=rate_std, size=l2_im.data.shape
    ).astype(np.float32)

    l2_im.meta["wcs"] = mk_gwcs(image_shape, sky_offset=sky_offset, rotate=rotation)

    l2_im.meta["background"] = mk_sky_background(
        level=None, subtracted=False, method="None"
    )

    l2_im.meta.cal_step["skymatch"] = "INCOMPLETE"
    return l2_im


@pytest.fixture
def wfi_rate():
    return mk_image_model()


@pytest.fixture
def mk_sky_match_image_models():
    rng = np.random.default_rng(1)
    im1a = mk_image_model(rng=rng)
    im1b = mk_image_model(sky_offset=[2, 2] * u.arcsec)
    im2a = mk_image_model(rotation=30 * u.deg)
    im2b = mk_image_model(sky_offset=[4, 4] * u.arcsec, rotation=60 * u.deg)
    im3 = mk_image_model(rotation=60 * u.deg)

    # add "bad" data
    im1a, dq_mask = _add_bad_pixels(im1a, 1e6, 1e9)
    im1b, _ = _add_bad_pixels(im1b, 1e6, 1e9)
    im2a, _ = _add_bad_pixels(im2a, 5e6, 3e9)
    im2b, _ = _add_bad_pixels(im2b, 5e6, 3e9)
    im3, _ = _add_bad_pixels(im3, 7e6, 1e8)

    return [im1a, im1b, im2a, im2b, im3], dq_mask


def _add_bad_pixels(im, sat_val, dont_use_val):
    # Add two types of "bad" pixels: 1) in the center of the image that will
    # lie in the intersection region common to all images (we assume images
    # are rotated around a common center) and 2) bad pixels at the corners
    # of the images that have a different flag and will be excluded from the
    # analysis of rotated images because they will lie outside of the
    # intersection region common to all images (and not because they are
    # masked out based on DQ array).
    mask = np.ones(im.data.shape, dtype=bool)
    # Add some "bad" pixels:
    # corners
    im.data[:5, :5] = sat_val
    im.data[-5:, :5] = sat_val
    im.data[-5:, -5:] = sat_val
    im.data[:5, -5:] = sat_val

    im.dq[:5, :5] = pixel.SATURATED
    im.dq[-5:, :5] = pixel.SATURATED
    im.dq[-5:, -5:] = pixel.SATURATED
    im.dq[:5, -5:] = pixel.SATURATED

    mask[:5, :5] = False
    mask[-5:, :5] = False
    mask[-5:, -5:] = False
    mask[:5, -5:] = False

    cy, cx = (x // 2 for x in im.data.shape)
    cx -= 5
    cy -= 5

    # center
    im.data[cx : cx + 10, cy : cy + 10] = dont_use_val
    im.dq[cx : cx + 10, cy : cy + 10] = pixel.DO_NOT_USE
    mask[cx : cx + 10, cy : cy + 10] = False

    return im, mask


@pytest.mark.parametrize(
    "skymethod, subtract, skystat, match_down",
    tuple(
        product(
            ["local", "match", "global", "global+match"],
            [False, True],
            ["median", "mean", "midpt", "mode"],
            [False, True],
        )
    ),
)
def test_skymatch(wfi_rate, skymethod, subtract, skystat, match_down):
    # test basic functionality and correctness of sky computations
    rng = np.random.default_rng(42)
    im1 = wfi_rate.copy()
    im2 = im1.copy()
    im3 = im1.copy()

    # add "bad" data
    im1, dq_mask = _add_bad_pixels(im1, 1e6, 1e9)
    im2, _ = _add_bad_pixels(im2, 5e6, 3e9)
    im3, _ = _add_bad_pixels(im3, 7e6, 1e8)

    library = ModelLibrary([im1, im2, im3])

    # define some background:
    levels = [9.12, 8.28, 2.56]

    with library:
        for i, (im, lev) in enumerate(zip(library, levels)):
            im.data = rng.normal(loc=lev, scale=0.05, size=im.data.shape)
            library.shelve(im, i)

    # exclude central DO_NOT_USE and corner SATURATED pixels
    result = SkyMatchStep.call(
        library,
        skymethod=skymethod,
        match_down=match_down,
        subtract=subtract,
        skystat=skystat,
        binwidth=0.2,
        nclip=0,
        dqbits="~DO_NOT_USE+SATURATED",
    )

    if skymethod in ["local", "global+match"]:
        ref_levels = levels

    elif skymethod == "match":
        lev0 = min(levels) if match_down else max(levels)
        ref_levels = np.array(levels) - lev0

    elif skymethod == "global":
        ref_levels = len(levels) * [min(levels)]

    sub_levels = np.array(levels) - np.array(ref_levels)

    with result:
        for i, (im, lev, rlev, slev) in enumerate(
            zip(result, levels, ref_levels, sub_levels)
        ):
            # check that meta was set correctly:
            assert im.meta.background.method == skymethod
            assert im.meta.background.subtracted == subtract

            # test computed/measured sky values if level is set:
            if not np.isclose(im.meta.background.level, 0):
                assert abs(im.meta.background.level - rlev) < 0.01

            # test
            if subtract:
                assert abs(np.mean(im.data[dq_mask]) - slev) < 0.01
            else:
                assert abs(np.mean(im.data[dq_mask]) - lev) < 0.01
            result.shelve(im, i, modify=False)


@pytest.mark.parametrize(
    "skymethod, subtract, skystat",
    tuple(product(["local", "match", "global"], [False, True], ["mean"])),
)
def test_skymatch_overlap(mk_sky_match_image_models, skymethod, subtract, skystat):
    # test that computations are performed only in the area of overlap
    # between images (bad pixels in the corners of rotated images are ignored).
    # Set 'nclip' to 0 in order to not clip bad pixels in computing mean.
    rng = np.random.default_rng(7)
    [im1a, im1b, im2a, im2b, im3], dq_mask = mk_sky_match_image_models

    library = ModelLibrary([im1a, im1b, im2a, im2b, im3])

    # define some background:
    levels = [9.12, 9.12, 8.28, 8.28, 2.56]

    with library:
        for i, (im, lev) in enumerate(zip(library, levels)):
            im.data = rng.normal(loc=lev, scale=0.01, size=im.data.shape)
            library.shelve(im, i)

    # We do not exclude SATURATED pixels. They should be ignored because
    # images are rotated and SATURATED pixels in the corners are not in the
    # common intersection of all input images. This is the purpose of this test
    result = SkyMatchStep.call(
        library,
        skymethod=skymethod,
        match_down=True,
        subtract=subtract,
        skystat=skystat,
        nclip=0,
        dqbits="~DO_NOT_USE",  # specifically DO NOT add 'SATURATED' flag
    )

    if skymethod in ["local", "global+match"]:
        ref_levels = levels

    elif skymethod == "match":
        ref_levels = np.array(levels) - min(levels)

    elif skymethod == "global":
        ref_levels = len(levels) * [min(levels)]

    sub_levels = np.array(levels) - np.array(ref_levels)

    with result:
        for i, (im, lev, rlev, slev) in enumerate(
            zip(result, levels, ref_levels, sub_levels)
        ):
            # check that meta was set correctly:
            assert im.meta.background.method == skymethod
            assert im.meta.background.subtracted == subtract

            if skymethod in ["local", "global"]:
                # These two sky methods must fail because they do not take
                # into account (do not compute) overlap regions and use
                # entire images:
                assert abs(im.meta.background.level - rlev) < 0.1

                # test
                if subtract:
                    assert abs(np.mean(im.data[dq_mask]) - slev) < 0.1
                else:
                    assert abs(np.mean(im.data[dq_mask]) - lev) < 0.01
            else:
                # test computed/measured sky values if level is nonzero:
                if not np.isclose(im.meta.background.level, 0):
                    assert abs(im.meta.background.level - rlev) < 0.01

                # test
                if subtract:
                    assert abs(np.mean(im.data[dq_mask]) - slev) < 0.01
                else:
                    assert abs(np.mean(im.data[dq_mask]) - lev) < 0.01
            result.shelve(im, i, modify=False)


@pytest.mark.parametrize(
    "skymethod, subtract",
    tuple(product(["local", "match", "global", "global+match"], [False, True])),
)
def test_skymatch_2x(wfi_rate, skymethod, subtract):
    # Test that repetitive applications of skymatch produce the same results
    rng = np.random.default_rng(19)
    im1 = wfi_rate.copy()
    im2 = im1.copy()
    im3 = im1.copy()

    # add "bad" data
    im1, dq_mask = _add_bad_pixels(im1, 1e6, 1e9)
    im2, _ = _add_bad_pixels(im2, 5e6, 3e9)
    im3, _ = _add_bad_pixels(im3, 7e6, 1e8)

    library = ModelLibrary([im1, im2, im3])

    # define some background:
    levels = [9.12, 8.28, 2.56]

    with library:
        for i, (im, lev) in enumerate(zip(library, levels)):
            im.data = rng.normal(loc=lev, scale=0.05, size=im.data.shape)
            library.shelve(im, i)

    # We do not exclude SATURATED pixels. They should be ignored because
    # images are rotated and SATURATED pixels in the corners are not in the
    # common intersection of all input images. This is the purpose of this test
    step = SkyMatchStep(
        skymethod=skymethod,
        match_down=True,
        subtract=subtract,
        skystat="mean",
        nclip=0,
        dqbits="~DO_NOT_USE+SATURATED",
    )

    result = step.run([im1, im2, im3])

    with result:
        model = result.borrow(0)
        assert model.meta.background.subtracted == step.subtract
        assert model.meta.background.level is not None
        result.shelve(model, 0, modify=False)

    # 2nd run.
    step.subtract = False
    result2 = step.run(result)

    with result2:
        model = result2.borrow(0)
        assert model.meta.background.subtracted == step.subtract
        assert model.meta.background.level is not None
        result2.shelve(model, 0, modify=False)

    # compute expected levels
    if skymethod in ["local", "global+match"]:
        ref_levels = levels

    elif skymethod == "match":
        ref_levels = np.array(levels) - min(levels)

    elif skymethod == "global":
        ref_levels = len(levels) * [min(levels)]

    sub_levels = np.array(levels) - np.array(ref_levels)

    # compare results
    with result2:
        for i, (im, lev, rlev, slev) in enumerate(
            zip(result2, levels, ref_levels, sub_levels)
        ):
            # check that meta was set correctly:
            assert im.meta.background.method == skymethod
            assert im.meta.background.subtracted == step.subtract

            # test computed/measured sky values:
            if not np.isclose(im.meta.background.level, 0, atol=1e-6):
                assert abs(im.meta.background.level - rlev) < 0.01

            # test
            if subtract:
                assert abs(np.mean(im.data[dq_mask]) - slev) < 0.01
            else:
                assert abs(np.mean(im.data[dq_mask]) - lev) < 0.01
            result2.shelve(im, i, modify=False)


@pytest.mark.parametrize(
    "input_type",
    [
        "ModelLibrary",
        "ASNFile",
        "DataModelList",
        "ASDFFilenameList",
    ],
)
def test_skymatch_always_returns_modellibrary_with_updated_datamodels(
    input_type,
    mk_sky_match_image_models,
    tmp_path,
    create_mock_asn_file,
):
    """Test that the SkyMatchStep always returns a ModelLibrary
    with updated data models after processing different input types."""

    os.chdir(tmp_path)
    [im1a, im1b, im2a, im2b, im3], dq_mask = mk_sky_match_image_models

    im1a.meta.filename = "im1a.asdf"
    im1b.meta.filename = "im1b.asdf"
    im2a.meta.filename = "im2a.asdf"
    im2b.meta.filename = "im2b.asdf"
    im3.meta.filename = "im3.asdf"

    library = ModelLibrary([im1a, im1b, im2a, im2b, im3])
    library._save(tmp_path)

    step_input_map = {
        "ModelLibrary": library,
        "ASNFile": create_mock_asn_file(
            tmp_path,
            members_mapping=[
                {"expname": im1a.meta.filename, "exptype": "science"},
                {"expname": im1b.meta.filename, "exptype": "science"},
                {"expname": im2a.meta.filename, "exptype": "science"},
                {"expname": im2b.meta.filename, "exptype": "science"},
                {"expname": im3.meta.filename, "exptype": "science"},
            ],
        ),
        "DataModelList": [im1a, im1b],
        "ASDFFilenameList": [im1a.meta.filename, im1b.meta.filename],
    }

    step_input = step_input_map.get(input_type)

    res = SkyMatchStep.call(step_input)

    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert model.meta.cal_step.skymatch == "COMPLETE"
            assert hasattr(model.meta, "background")
            res.shelve(model, i, modify=False)



import functools
import logging
from typing import List, Tuple, Union

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.modeling import models as astmodels
from astropy.utils.misc import isiterable
from gwcs import WCS
from gwcs.wcstools import wcs_from_fiducial
from roman_datamodels.datamodels import DataModel
from stcal.alignment.util import compute_s_region_keyword

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

_MAX_SIP_DEGREE = 6


def wcs_bbox_from_shape(shape):
    """Create a bounding box from the shape of the data.
    This is appropriate to attach to a wcs object.

    Parameters
    ----------
    shape : tuple
        The shape attribute from a `numpy.ndarray` array.

    Returns
    -------
    bbox : tuple
        Bounding box in x, y order.
    """
    bbox = ((-0.5, shape[-1] - 0.5), (-0.5, shape[-2] - 0.5))
    return bbox


def wcs_from_footprints(
    dmodels,
    refmodel=None,
    transform=None,
    bounding_box=None,
    pscale_ratio=None,
    pscale=None,
    rotation=None,
    shape=None,
    ref_pixel: Tuple[float, float] = None,
    ref_coord: Tuple[float, float] = None,
):
    """
    Create a WCS from a list of input data models.

    A fiducial point in the output coordinate frame is created from  the
    footprints of all WCS objects. For a spatial frame this is the center
    of the union of the footprints. For a spectral frame the fiducial is in
    the beginning of the footprint range.
    If ``refmodel`` is None, the first WCS object in the list is considered
    a reference. The output coordinate frame and projection (for celestial frames)
    is taken from ``refmodel``.
    If ``transform`` is not supplied, a compound transform is created using
    CDELTs and PC.
    If ``bounding_box`` is not supplied, the bounding_box of the new WCS is computed
    from bounding_box of all input WCSs.

    Parameters
    ----------
    dmodels : list of `~jwst.datamodels.DataModel`
        A list of data models.
    refmodel : `~jwst.datamodels.DataModel`, optional
        This model's WCS is used as a reference.
        WCS. The output coordinate frame, the projection and a
        scaling and rotation transform is created from it. If not supplied
        the first model in the list is used as ``refmodel``.
    transform : `~astropy.modeling.core.Model`, optional
        A transform, passed to :meth:`~gwcs.wcstools.wcs_from_fiducial`
        If not supplied Scaling | Rotation is computed from ``refmodel``.
    bounding_box : tuple, optional
        Bounding_box of the new WCS.
        If not supplied it is computed from the bounding_box of all inputs.
    pscale_ratio : float, None, optional
        Ratio of input to output pixel scale. Ignored when either
        ``transform`` or ``pscale`` are provided.
    pscale : float, None, optional
        Absolute pixel scale in degrees. When provided, overrides
        ``pscale_ratio``. Ignored when ``transform`` is provided.
    rotation : float, None, optional
        Position angle (in degrees) of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes. Ignored when ``transform`` is
        provided.
    shape : tuple of int, None, optional
        Shape of the image (data array) using ``numpy.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned to
        ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.
    ref_pixel : tuple of float, None, optional
        Position of the reference pixel in the image array.  If ``ref_pixel`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.
    ref_coord : tuple of float, None, optional
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    """
    bb = bounding_box
    wcslist = [im.meta.wcs for im in dmodels]

    if not isiterable(wcslist):
        raise ValueError("Expected 'wcslist' to be an iterable of WCS objects.")

    if not all([isinstance(w, WCS) for w in wcslist]):
        raise TypeError("All items in wcslist are to be instances of gwcs.WCS.")

    if refmodel is None:
        refmodel = dmodels[0]
    else:
        if not isinstance(refmodel, DataModel):
            raise TypeError("Expected refmodel to be an instance of DataModel.")

    fiducial = compute_fiducial(wcslist, bb)
    if ref_coord is not None:
        # overwrite spatial axes with user-provided ref_coord:
        i = 0
        for k, axt in enumerate(wcslist[0].output_frame.axes_type):
            if axt == "SPATIAL":
                fiducial[k] = ref_coord[i]
                i += 1

    ref_fiducial = np.array(
        [refmodel.meta.wcsinfo.ra_ref, refmodel.meta.wcsinfo.dec_ref]
    )

    prj = astmodels.Pix2Sky_TAN()

    if transform is None:
        transform = []
        sky_axes = refmodel.meta.wcs._get_axes_indices().tolist()

        # Need to put the rotation matrix (List[float, float, float, float])
        # returned from calc_rotation_matrix into the correct shape for
        # constructing the transformation
        v3yangle = np.deg2rad(refmodel.meta.wcsinfo.v3yangle)
        vparity = refmodel.meta.wcsinfo.vparity
        if rotation is None:
            roll_ref = np.deg2rad(refmodel.meta.wcsinfo.roll_ref)
        else:
            roll_ref = np.deg2rad(rotation) + (vparity * v3yangle)

        pc = np.reshape(
            calc_rotation_matrix(roll_ref, v3yangle, vparity=vparity), (2, 2)
        )

        rotation = astmodels.AffineTransformation2D(pc, name="pc_rotation_matrix")
        transform.append(rotation)

        if sky_axes:
            if not pscale:
                pscale = compute_scale(
                    refmodel.meta.wcs, ref_fiducial, pscale_ratio=pscale_ratio
                )
            transform.append(
                astmodels.Scale(pscale, name="cdelt1")
                & astmodels.Scale(pscale, name="cdelt2")
            )

        if transform:
            transform = functools.reduce(lambda x, y: x | y, transform)

    out_frame = refmodel.meta.wcs.output_frame
    input_frame = refmodel.meta.wcs.input_frame
    wnew = wcs_from_fiducial(
        fiducial,
        coordinate_frame=out_frame,
        projection=prj,
        transform=transform,
        input_frame=input_frame,
    )

    footprints = [w.footprint().T for w in wcslist]
    domain_bounds = np.hstack([wnew.backward_transform(*f) for f in footprints])
    axis_min_values = np.min(domain_bounds, axis=1)
    domain_bounds = (domain_bounds.T - axis_min_values).T

    output_bounding_box = []
    for axis in out_frame.axes_order:
        axis_min, axis_max = (
            domain_bounds[axis].min(),
            domain_bounds[axis].max(),
        )
        output_bounding_box.append((axis_min, axis_max))

    output_bounding_box = tuple(output_bounding_box)
    if ref_pixel is None:
        offset1, offset2 = wnew.backward_transform(*fiducial)
        offset1 -= axis_min_values[0]
        offset2 -= axis_min_values[1]
    else:
        offset1, offset2 = ref_pixel
    offsets = astmodels.Shift(-offset1, name="ref_pixel1") & astmodels.Shift(
        -offset2, name="ref_pixel2"
    )

    wnew.insert_transform("detector", offsets, after=True)
    wnew.bounding_box = output_bounding_box

    if shape is None:
        shape = [int(axs[1] - axs[0] + 0.5) for axs in output_bounding_box[::-1]]

    wnew.pixel_shape = shape[::-1]
    wnew.array_shape = shape

    return wnew


def compute_scale(
    wcs: WCS,
    fiducial: Union[tuple, np.ndarray],
    disp_axis: int = None,
    pscale_ratio: float = None,
) -> float:
    """Compute scaling transform.

    Parameters
    ----------
    wcs : `~gwcs.wcs.WCS`
        Reference WCS object from which to compute a scaling factor.

    fiducial : tuple
        Input fiducial of (RA, DEC) or (RA, DEC, Wavelength) used in calculating
        reference points.

    disp_axis : int
        Dispersion axis integer. Assumes the same convention as
        `wcsinfo.dispersion_direction`

    pscale_ratio : int
        Ratio of input to output pixel scale

    Returns
    -------
    scale : float
        Scaling factor for x and y or cross-dispersion direction.

    """
    spectral = "SPECTRAL" in wcs.output_frame.axes_type

    if spectral and disp_axis is None:
        raise ValueError("If input WCS is spectral, a disp_axis must be given")

    crpix = np.array(wcs.invert(*fiducial))

    delta = np.zeros_like(crpix)
    spatial_idx = np.where(np.array(wcs.output_frame.axes_type) == "SPATIAL")[0]
    delta[spatial_idx[0]] = 1

    crpix_with_offsets = np.vstack((crpix, crpix + delta, crpix + np.roll(delta, 1))).T
    crval_with_offsets = wcs(*crpix_with_offsets, with_bounding_box=False)

    coords = SkyCoord(
        ra=crval_with_offsets[spatial_idx[0]],
        dec=crval_with_offsets[spatial_idx[1]],
        unit="deg",
    )
    xscale = np.abs(coords[0].separation(coords[1]).value)
    yscale = np.abs(coords[0].separation(coords[2]).value)

    if pscale_ratio is not None:
        xscale *= pscale_ratio
        yscale *= pscale_ratio

    if spectral:
        # Assuming scale doesn't change with wavelength
        # Assuming disp_axis is consistent with
        # DataModel.meta.wcsinfo.dispersion.direction
        return yscale if disp_axis == 1 else xscale

    return np.sqrt(xscale * yscale)


def calc_rotation_matrix(
    roll_ref: float, v3i_yang: float, vparity: int = 1
) -> List[float]:
    """Calculate the rotation matrix.

    Parameters
    ----------
    roll_ref : float
        Telescope roll angle of V3 North over East at the ref. point in radians

    v3i_yang : float
        The angle between ideal Y-axis and V3 in radians.

    vparity : int
        The x-axis parity, usually taken from the JWST SIAF parameter VIdlParity.
        Value should be "1" or "-1".

    Returns
    -------
    matrix: [pc1_1, pc1_2, pc2_1, pc2_2]
        The rotation matrix

    Notes
    -----
    The rotation is

       ----------------
       | pc1_1  pc2_1 |
       | pc1_2  pc2_2 |
       ----------------

    """
    if vparity not in (1, -1):
        raise ValueError(f"vparity should be 1 or -1. Input was: {vparity}")

    rel_angle = roll_ref - (vparity * v3i_yang)

    pc1_1 = vparity * np.cos(rel_angle)
    pc1_2 = np.sin(rel_angle)
    pc2_1 = vparity * -np.sin(rel_angle)
    pc2_2 = np.cos(rel_angle)

    return [pc1_1, pc1_2, pc2_1, pc2_2]


def compute_fiducial(wcslist, bounding_box=None):
    """
    For a celestial footprint this is the center.
    For a spectral footprint, it is the beginning of the range.

    This function assumes all WCSs have the same output coordinate frame.
    """

    axes_types = wcslist[0].output_frame.axes_type
    spatial_axes = np.array(axes_types) == "SPATIAL"
    spectral_axes = np.array(axes_types) == "SPECTRAL"
    footprints = np.hstack([w.footprint(bounding_box=bounding_box).T for w in wcslist])
    spatial_footprint = footprints[spatial_axes]
    spectral_footprint = footprints[spectral_axes]

    fiducial = np.empty(len(axes_types))
    if spatial_footprint.any():
        lon, lat = spatial_footprint
        lon, lat = np.deg2rad(lon), np.deg2rad(lat)
        x = np.cos(lat) * np.cos(lon)
        y = np.cos(lat) * np.sin(lon)
        z = np.sin(lat)

        x_mid = (np.max(x) + np.min(x)) / 2.0
        y_mid = (np.max(y) + np.min(y)) / 2.0
        z_mid = (np.max(z) + np.min(z)) / 2.0
        lon_fiducial = np.rad2deg(np.arctan2(y_mid, x_mid)) % 360.0
        lat_fiducial = np.rad2deg(np.arctan2(z_mid, np.sqrt(x_mid**2 + y_mid**2)))
        fiducial[spatial_axes] = lon_fiducial, lat_fiducial
    if spectral_footprint.any():
        fiducial[spectral_axes] = spectral_footprint.min()
    return fiducial


def create_footprint(wcs, shape=None, center=True):
    """Calculate sky footprint

    Parameters
    ----------
    wcs : `gwcs.WCS`
        The WCS information to get the footprint from

    shape : n-tuple or None
       Shape to use if wcs has no defined shape.

    center : bool
        If True use the center of the pixel, otherwise use the corner.

    Returns
    -------
    footprint : `numpy.ndarray`
        The footprint.
    """
    bbox = wcs.bounding_box

    if bbox is None:
        bbox = wcs_bbox_from_shape(shape)

    # footprint is an array of shape (2, 4) - i.e. 4 values for RA and 4 values for
    # Dec - as we are interested only in the footprint on the sky
    footprint = wcs.footprint(bbox, center=center, axis_type="spatial").T
    # take only imaging footprint
    footprint = footprint[:2, :]

    # Make sure RA values are all positive
    negative_ind = footprint[0] < 0
    if negative_ind.any():
        footprint[0][negative_ind] = 360 + footprint[0][negative_ind]

    footprint = footprint.T
    return footprint


def add_s_region(model):
    """
    Calculate the detector's footprint using ``WCS.footprint`` and save it in the
    ``S_REGION`` keyword

    Parameters
    ----------
    model : `~roman_datamodels.datamodels.ImageModel`
        The data model for processing

    Returns
    -------
    A formatted string representing the detector's footprint
    """
    update_s_region_keyword(model, create_footprint(model.meta.wcs, shape=model.shape))


def update_s_region_keyword(model, footprint):
    s_region = compute_s_region_keyword(footprint)
    log.info(f"S_REGION VALUES: {s_region}")
    if "nan" in s_region:
        # do not update s_region if there are NaNs.
        log.info("There are NaNs in s_region, S_REGION not updated.")
    else:
        model.meta.wcsinfo.s_region = s_region
        log.info(f"Update S_REGION to {model.meta.wcsinfo.s_region}")


def list_1d_to_2d(l, n):
    """Convert 1-dimensional list to 2-dimensional

    Parameters
    ----------
    l : list
        The list to convert.

    n : int
       The length of the x dimension, or the length of the inner lists.

    Returns
    -------
    l2d : list of lists
        The 2D form
    """
    return [l[i : i + n] for i in range(0, len(l), n)]



"""
Utilities
"""

import inspect
import logging
from importlib import import_module
from pkgutil import walk_packages

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Step classes that are not user-api steps
NON_STEPS = [
    "EngDBLogStep",
    "FunctionWrapper",
    "RomancalPipeline",
    "RomanPipeline",
    "ExposurePipeline",
    "MosaicPipeline",
    "RomanStep",
    "Pipeline",
    "Step",
    "SystemCall",
]


def all_steps():
    """List all classes subclassed from Step

    Returns
    -------
    steps : dict
        Key is the classname, value is the class
    """
    from romancal.stpipe import RomanStep as Step

    romancal = import_module("romancal")

    steps = {}
    for module in load_sub_modules(romancal):
        more_steps = {
            klass_name: klass
            for klass_name, klass in inspect.getmembers(
                module, lambda o: inspect.isclass(o) and issubclass(o, Step)
            )
            if klass_name not in NON_STEPS
        }
        steps.update(more_steps)

    return steps


def load_sub_modules(module):
    """
    Recursively loads all submodules of a module (this is not a local import).

    Parameters
    ----------
    module : module
        A python module to walk, load

    Returns
    -------
    generator
        A generator of all submodules of module recursively until no more sub modules are found
    """

    for package_info in walk_packages(module.__path__):
        if package_info.module_finder.path.startswith(module.__path__[0]):
            package = import_module(f"{module.__name__}.{package_info.name}")

            if package_info.ispkg:
                yield from load_sub_modules(package)

            yield package



import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def save_median(example_model, median_data, median_wcs, make_output_path):
    _save_intermediate_output(
        _make_median_model(example_model, median_data, median_wcs),
        "median",
        make_output_path,
    )


def save_drizzled(drizzled_model, make_output_path):
    _save_intermediate_output(drizzled_model, "outlier_i2d", make_output_path)


def _make_median_model(example_model, data, wcs):
    model = example_model.copy()
    model.data = data
    model.meta.filename = "drizzled_median.asdf"
    model.meta.wcs = wcs
    return model


def _save_intermediate_output(model, suffix, make_output_path):
    """
    Ensure all intermediate outputs from OutlierDetectionStep have consistent file naming conventions

    Notes
    -----
    self.make_output_path() is updated globally for the step in the main pipeline
    to include the asn_id in the output path, so no need to handle it here.
    """

    # outlier_?2d is not a known suffix, and make_output_path cannot handle an
    # underscore in an unknown suffix, so do a manual string replacement
    input_path = model.meta.filename.replace("_outlier_", "_")

    output_path = make_output_path(input_path, suffix=suffix)
    model.save(output_path)
    log.info(f"Saved {suffix} model in {output_path}")



import numpy as np
import pytest
from astropy.time import Time
from roman_datamodels import maker_utils, stnode
from roman_datamodels.datamodels import FlatRefModel, ImageModel

from romancal.flatfield import FlatFieldStep

RNG = np.random.default_rng(172)


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_flatfield_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a FLAT reffile"""

    shape = (20, 20)

    wfi_image = maker_utils.mk_level2_image(shape=shape)
    wfi_image.meta.instrument.name = instrument
    wfi_image.meta.instrument.detector = "WFI01"
    wfi_image.meta.instrument.optical_element = "F158"
    wfi_image.meta.exposure.type = exptype
    wfi_image.data = np.ones(shape, dtype=np.float32)
    wfi_image.dq = np.zeros(shape, dtype=np.uint32)
    wfi_image.err = np.zeros(shape, dtype=np.float32)
    wfi_image.var_poisson = np.zeros(shape, dtype=np.float32)
    wfi_image.var_rnoise = np.zeros(shape, dtype=np.float32)
    wfi_image.var_flat = np.zeros(shape, dtype=np.float32)

    wfi_image_model = ImageModel(wfi_image)
    flatref = stnode.FlatRef()
    meta = maker_utils.mk_ref_common("FLAT")
    meta["instrument"]["optical_element"] = "F158"
    meta["instrument"]["detector"] = "WFI01"
    flatref["meta"] = meta
    flatref["data"] = np.ones(shape, dtype=np.float32)
    flatref["dq"] = np.zeros(shape, dtype=np.uint32)
    flatref["err"] = (RNG.uniform(size=shape) * 0.05).astype(np.float32)
    flatref_model = FlatRefModel(flatref)

    result = FlatFieldStep.call(wfi_image_model, override_flat=flatref_model)

    assert (result.data == wfi_image.data).all()
    assert result.var_flat.shape == shape
    assert result.meta.cal_step.flat_field == "COMPLETE"

    # test that the step is skipped if the reference file is N/A
    result = FlatFieldStep.call(wfi_image_model, override_flat="N/A")

    assert result.meta.cal_step.flat_field == "SKIPPED"


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_crds_temporal_match(instrument, exptype):
    """Test that the basic inferface works for data requiring a FLAT reffile"""

    wfi_image = maker_utils.mk_level2_image()
    wfi_image.meta.instrument.name = instrument
    wfi_image.meta.instrument.detector = "WFI01"
    wfi_image.meta.instrument.optical_element = "F158"

    wfi_image.meta.exposure.start_time = Time("2020-01-02T11:11:11.110")
    wfi_image.meta.exposure.end_time = Time("2020-01-02T11:33:11.110")

    wfi_image.meta.exposure.type = exptype
    wfi_image_model = ImageModel(wfi_image)

    step = FlatFieldStep()
    ref_file_path = step.get_reference_file(wfi_image_model, "flat")

    wfi_image_model.meta.exposure.start_time = Time("2021-09-01T00:11:11.110")
    wfi_image_model.meta.exposure.end_time = Time("2021-09-01T00:33:11.110")
    ref_file_path_b = step.get_reference_file(wfi_image_model, "flat")
    assert "/".join(ref_file_path.rsplit("/", 1)[1:]) != "/".join(
        ref_file_path_b.rsplit("/", 1)[1:]
    )


@pytest.mark.parametrize(
    "instrument",
    [
        "WFI",
    ],
)
@pytest.mark.parametrize(
    "exptype",
    [
        "WFI_GRISM",
        "WFI_PRISM",
    ],
)
def test_spectroscopic_skip(instrument, exptype):
    wfi_image = maker_utils.mk_level2_image()
    wfi_image.meta.instrument.name = instrument
    wfi_image.meta.instrument.detector = "WFI01"
    wfi_image.meta.instrument.optical_element = "F158"

    wfi_image.meta.exposure.start_time = Time("2020-02-01T00:00:00.000")
    wfi_image.meta.exposure.end_time = Time("2020-02-01T00:00:05.000")

    wfi_image.meta.exposure.type = exptype
    wfi_image_model = ImageModel(wfi_image)

    result = FlatFieldStep.call(wfi_image_model)
    assert result.meta.cal_step.flat_field == "SKIPPED"



import pytest

from romancal.refpix import RefPixStep
from romancal.refpix.refpix import run_steps


@pytest.mark.parametrize(
    "remove_offset, remove_trends, cosine_interpolate, fft_interpolate",
    [(True, True, True, False), (True, True, True, True)],
)
def test_refpix_step(
    datamodel,
    ref_pix_ref,
    remove_offset,
    remove_trends,
    cosine_interpolate,
    fft_interpolate,
):
    # Setup and run the correction outside of the step, this is
    # already regression tested.
    regression_datamodel = datamodel.copy()
    regression_ref_pix_ref = ref_pix_ref.copy()
    regression = run_steps(
        regression_datamodel,
        regression_ref_pix_ref,
        remove_offset,
        remove_trends,
        cosine_interpolate,
        fft_interpolate,
    )

    # Run the step
    result = RefPixStep.call(
        datamodel,
        override_refpix=ref_pix_ref,
        remove_offset=remove_offset,
        remove_trends=remove_trends,
        cosine_interpolate=cosine_interpolate,
        fft_interpolate=fft_interpolate,
    )

    # Check the outputs are distinct
    assert result is not regression

    # Check the data
    assert (result.data == regression.data).all()
    # Check the amp33
    assert (result.amp33 == regression.amp33).all()
    # Check left ref pix
    assert (result.border_ref_pix_left == regression.border_ref_pix_left).all()
    # Check right ref pix
    assert (result.border_ref_pix_right == regression.border_ref_pix_right).all()
    #
    # Run the step with reffile = N/A
    result = RefPixStep.call(
        datamodel,
        override_refpix="N/A",
        remove_offset=remove_offset,
        remove_trends=remove_trends,
        cosine_interpolate=cosine_interpolate,
        fft_interpolate=fft_interpolate,
    )
    assert result.meta.cal_step.refpix == "SKIPPED"



from romancal.associations import AssociationPool
from romancal.associations.tests.helpers import t_path

POOL_FILE = t_path("data/jw93060_20150312T160130_pool.csv")


def test_pool(tmp_path):
    pool = AssociationPool.read(POOL_FILE)
    assert len(pool) == 636

    file_path = tmp_path / __name__
    file_path.mkdir()

    tmp_pool = str(file_path / "tmp_pool.csv")
    pool.write(tmp_pool)

    roundtrip = AssociationPool.read(tmp_pool)
    assert len(pool) == len(roundtrip)
    assert set(pool.colnames) == set(roundtrip.colnames)



import numpy as np
import pytest
from asdf import AsdfFile
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from gwcs import WCS
from gwcs import coordinate_frames as cf
from roman_datamodels import datamodels, maker_utils

from romancal.resample import ResampleStep, resample_utils


class MockModel:
    def __init__(self, pixel_area, pixel_scale_ratio):
        self.meta = MockMeta(pixel_area, pixel_scale_ratio)


class MockMeta:
    def __init__(self, pixel_area, pixel_scale_ratio):
        self.photometry = MockPhotometry(pixel_area)
        self.resample = MockResample(pixel_scale_ratio)


class MockPhotometry:
    def __init__(self, pixel_area):
        self.pixel_area = pixel_area


class MockResample:
    def __init__(self, pixel_scale_ratio):
        self.pixel_scale_ratio = pixel_scale_ratio


class Mosaic:
    def __init__(self, fiducial_world, pscale, shape, filename, n_images):
        self.fiducial_world = fiducial_world
        self.pscale = pscale
        self.shape = shape
        self.filename = filename
        self.n_images = n_images

    def create_mosaic(self):
        """
        Create a dummy L3 datamodel given the coordinates of the fiducial point,
        a pixel scale, and the image shape and filename.

        Returns
        -------
        datamodels.MosaicModel
            An L3 MosaicModel datamodel.
        """
        l3 = maker_utils.mk_level3_mosaic(
            shape=self.shape,
            n_images=self.n_images,
        )
        # data from WFISim simulation of SCA #01
        l3.meta.filename = self.filename
        l3.meta["wcs"] = create_wcs_object_without_distortion(
            fiducial_world=self.fiducial_world,
            pscale=self.pscale,
            shape=self.shape,
        )
        l3.meta.wcs.forward_transform
        return datamodels.MosaicModel(l3)


def create_wcs_object_without_distortion(fiducial_world, pscale, shape, **kwargs):
    """
    Create a simple WCS object without either distortion or rotation.

    Parameters
    ----------
    fiducial_world : tuple
        A pair of values corresponding to the fiducial's world coordinate.
    pscale : tuple
        A pair of values corresponding to the pixel scale in each axis.
    shape : tuple
        A pair of values specifying the dimensions of the WCS object.

    Returns
    -------
    gwcs.WCS
        A gwcs.WCS object.
    """
    # components of the model
    shift = models.Shift() & models.Shift()
    affine = models.AffineTransformation2D(
        matrix=[[1, 0], [0, 1]], translation=[0, 0], name="pc_rotation_matrix"
    )
    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    # transforms between frames
    # detector -> sky
    det2sky = shift | affine | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector",
        axes_order=(0, 1),
        axes_names=("x", "y"),
        unit=(u.pix, u.pix),
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg)
    )

    pipeline = [
        (detector_frame, det2sky),
        (sky_frame, None),
    ]

    wcs_obj = WCS(pipeline)

    wcs_obj.bounding_box = kwargs.get(
        "bounding_box",
        (
            (-0.5, shape[-1] - 0.5),
            (-0.5, shape[-2] - 0.5),
        ),
    )

    wcs_obj.pixel_shape = shape[::-1]
    wcs_obj.array_shape = shape

    return wcs_obj


@pytest.fixture
def asdf_wcs_file():
    def _create_asdf_wcs_file(tmp_path):
        file_path = tmp_path / "wcs.asdf"
        wcs_data = create_wcs_object_without_distortion(
            (10, 0),
            (0.000031, 0.000031),
            (100, 100),
        )
        wcs = {"wcs": wcs_data}
        with AsdfFile(wcs) as af:
            af.write_to(file_path)
        return str(file_path)

    return _create_asdf_wcs_file


@pytest.mark.parametrize(
    "vals, name, min_vals, expected",
    [
        ([1, 2], "list1", None, [1, 2]),
        ([None, None], "list2", None, None),
        ([1, 2], "list4", [0, 0], [1, 2]),
    ],
)
def test_check_list_pars_valid(vals, name, min_vals, expected):
    step = ResampleStep()

    result = step._check_list_pars(vals, name, min_vals)
    assert result == expected


def test_load_custom_wcs_no_file():
    step = ResampleStep()
    result = step._load_custom_wcs(None, (512, 512))
    assert result is None


def test_load_custom_wcs_missing_output_shape(asdf_wcs_file):
    with pytest.raises(ValueError):
        step = ResampleStep()
        step._load_custom_wcs(asdf_wcs_file, None)


def test_load_custom_wcs_invalid_file(tmp_path):
    step = ResampleStep()
    invalid_file = tmp_path / "invalid.asdf"
    with open(invalid_file, "w") as f:
        f.write("invalid asdf file")

    with pytest.raises(ValueError):
        step._load_custom_wcs(str(invalid_file), (512, 512))


def test_load_custom_wcs_asdf_without_wcs_attribute(tmp_path):
    step = ResampleStep()
    file_path = tmp_path / "asdf_file.asdf"
    wcs = {}
    with AsdfFile(wcs) as af:
        af.write_to(file_path)

    with pytest.raises(KeyError):
        step._load_custom_wcs(str(file_path), (100, 100))


@pytest.mark.parametrize(
    "vals, name, min_vals, expected",
    [
        ([1, 2], "test", [0, 0], [1, 2]),
        ([None, None], "test", [0, 0], None),
        ([0, 0], "test", [0, 0], [0, 0]),
        ([1, 1], "test", [0, 0], [1, 1]),
        ([0, 1], "test", [0, 0], [0, 1]),
        ([1, 0], "test", [0, 0], [1, 0]),
    ],
)
def test_check_list_pars(vals, name, min_vals, expected):
    step = ResampleStep()

    result = step._check_list_pars(vals, name, min_vals)
    assert result == expected


@pytest.mark.parametrize(
    "vals, name, min_vals",
    [
        ([None, 2], "test", [0, 0]),
        ([1, None], "test", [0, 0]),
        ([1], "test", [0, 0]),
        ([1, 2, 3], "test", [0, 0]),
        ([None, None, None], "test", [0, 0]),
        ([1, 2], "test", [2, 2]),
    ],
)
def test_check_list_pars_exception(vals, name, min_vals):
    step = ResampleStep()

    with pytest.raises(ValueError):
        step._check_list_pars(vals, name, min_vals)


@pytest.mark.parametrize(
    """pixel_area, pixel_scale_ratio,
    expected_steradians, expected_arcsecsq""",
    [
        # Happy path tests
        (1.0, 2.0, 4.0, 4.0),
        (2.0, 0.5, 0.5, 0.5),
        (0.0, 2.0, 0.0, 0.0),
        (1.0, 0.0, 0.0, 0.0),
        (None, 2.0, None, 4.0),
        (1.0, 2.0, 4.0, None),
    ],
)
def test_update_phot_keywords(
    pixel_area,
    pixel_scale_ratio,
    expected_steradians,
    expected_arcsecsq,
):
    step = ResampleStep()
    model = MockModel(pixel_area, pixel_scale_ratio)

    step.update_phot_keywords(model)

    assert model.meta.photometry.pixel_area == expected_steradians


@pytest.mark.parametrize(
    "good_bits, dq_array, expected_output",
    [
        (
            "~DO_NOT_USE+NON_SCIENCE",
            np.array([[0, 2**0, 0], [0, 2**13, 0], [0, 2**9, 0]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 0, 1]]),
        ),
        (
            "~513",
            np.array([[0, 2**0, 0], [2**13, 0, 0], [0, 0, 2**9]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 1, 0]]),
        ),
        (
            "~1+512",
            np.array([[2**13, 0, 0], [0, 0, 2**9], [0, 2**0, 0]]),
            np.array([[1, 1, 1], [1, 1, 0], [1, 0, 1]]),
        ),
        (
            "~1,512",
            np.array([[2**13, 2**0, 2**9], [0, 0, 0], [0, 0, 0]]),
            np.array([[1, 0, 0], [1, 1, 1], [1, 1, 1]]),
        ),
        (
            "LOW_QE+NONLINEAR",
            np.array([[2**13, 2**0, 2**16], [0, 0, 0], [0, 0, 0]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 1, 1]]),
        ),
        (
            "73728",
            np.array([[0, 0, 0], [0, 2**13, 2**0], [0, 2**16, 0]]),
            np.array([[1, 1, 1], [1, 1, 0], [1, 1, 1]]),
        ),
        (
            "8192+65536",
            np.array([[0, 0, 0], [0, 2**13, 0], [2**0, 2**16, 0]]),
            np.array([[1, 1, 1], [1, 1, 1], [0, 1, 1]]),
        ),
        (
            "8192,65536",
            np.array([[0, 0, 0], [0, 2**13, 0], [0, 2**16, 2**0]]),
            np.array([[1, 1, 1], [1, 1, 1], [1, 1, 0]]),
        ),
    ],
)
def test_build_driz_weight_multiple_good_bits(
    base_image, good_bits, dq_array, expected_output
):
    data_shape = dq_array.shape
    img1 = base_image()
    img1.dq = dq_array
    img1.data = np.ones(data_shape)

    result = resample_utils.build_driz_weight(
        img1, weight_type=None, good_bits=good_bits
    )

    np.testing.assert_array_equal(result, expected_output)


@pytest.mark.parametrize(
    "good_bits",
    [
        "~DO_NOT_USE+NON_SCIENCE",
        "~513",
        "~1+512",
        "~1,512",
        "LOW_QE+NONLINEAR",
        "73728",
        "8192+65536",
        "8192,65536",
    ],
)
def test_set_good_bits_in_resample_meta(base_image, good_bits):
    model = maker_utils.mk_level2_image(shape=(100, 100))
    model.meta.wcsinfo["vparity"] = -1

    img = datamodels.ImageModel(model)

    img.data *= img.meta.photometry.conversion_megajanskys / img.data

    step = ResampleStep

    res = step.call(img, good_bits=good_bits)

    assert res.meta.resample.good_bits == good_bits


@pytest.mark.parametrize("weight_type", ["ivm", "exptime", None])
def test_build_driz_weight_different_weight_type(base_image, weight_type):
    rng = np.random.default_rng()
    img1 = base_image()
    # update attributes that will be used in building the weight array
    img1.meta.exposure.exposure_time = 10
    img1.var_rnoise = rng.normal(1, 0.1, size=img1.shape)
    # build the drizzle weight array
    result = resample_utils.build_driz_weight(
        img1, weight_type=weight_type, good_bits="~DO_NOT_USE+NON_SCIENCE"
    )

    expected_results = {
        "ivm": img1.var_rnoise**-1,
        "exptime": np.ones(img1.shape, dtype=img1.data.dtype)
        * img1.meta.exposure.exposure_time,
        None: np.ones(img1.shape, dtype=img1.data.dtype),
    }

    np.testing.assert_array_almost_equal(expected_results.get(weight_type), result)
